require 'amrita2/core'
require 'annotation'

module AmritaHandler
  module ViewComponent
    include Amrita2::REXMLShortCut
    VIEW_PREFIX = 'amview'
    @@registered = {}
    def register(ann, component)
      @@registered[ann] = component
    end
    def registered
      @@registered
    end
    module_function :register, :registered

    @@prototypes = {}
    def self.prototypes
      @@prototypes
    end
    #@@instances = []
    #def self.instances
    #  @@instances
    #end
    
    module ActionReplace
      def element_attr_str(method)
        "{ :name=> \"#{method}_" + '#{__id__}" }'
      end
    end
    
    class Base
      attr_reader :owner_class, :attr
      attr_accessor :owner_object
      def initialize(owner_class, attr)
        @owner_class = owner_class
        @attr = attr
      end
      def self.ann_name
        'ATTR_' + self.name.sub(/^.*::/, '').gsub(/[a-z][A-Z]/){|s| "#{s[0,1]}_#{s[1,1]}"}.upcase
      end
      def self.spec_option
        {}
      end

      def ann_name
        self.class.ann_name
      end

      def element_attr_str(method)
        { :name=>"#{method}" }.inspect
      end
	  def inject_parameter(param)
	    attr_name = if @attr.instance_of?(Hash)
          @attr[:name].to_s
        else
          @attr.to_s
        end
        do_inject_parameter(@owner_object, attr_name, param)
	  end
	  def do_inject_parameter(obj, attr, param)
        if obj.respond_to?("#{attr}=") and param[attr]
          puts "INJECTED #{attr} => id:#{obj.__id__}" if $DEBUG
          obj.__send__("#{attr}=", param[attr])
        end
	  end
	  
      def define_view_method
        define_view_method_body(@owner_class, @attr.to_s)
      end
      def define_view_method_body(clazz, method)
        clazz.module_eval %Q{
          def #{VIEW_PREFIX}_#{method}
            a(#{element_attr_str(method)}) { #{method} }
          end
        }
      end
      def self.register
        ann = ann_name.to_sym
        AmritaHandler::ViewComponent.register(ann, self)
        Annotation::ScanHandler.register(ann) do |owner_class, value|
          if AmritaHandler::ViewComponent.registered[ann]
            puts "Create Component Instance: #{ann}" if $DEBUG
            vc = self.new(owner_class, value)
            AmritaHandler::ViewComponent.prototypes[owner_class] ||= []
            AmritaHandler::ViewComponent.prototypes[owner_class] << vc
          end
        end
      end
    end
    
    class Value < Base
      def define_view_method_body(clazz, method)
        clazz.module_eval %Q{
          def #{VIEW_PREFIX}_#{method}
            a({:value=>#{method}}.update(#{element_attr_str(method)}))
          end
        }
      end
    end
    
    class ActionValue < Value
      include ActionReplace
    end

    class Param < Base
      #def define_view_method_body(clazz, method)
      #end
      register
    end
    
    class Hidden < Value
      register
    end

    class Link < Base
      def self.spec_option
        { :replace_attr => true,
          :link_attr => [:href, [/javascript:.*/]] }
      end
      def define_view_method
        p @attr
        method = @attr[:name].to_s
        href = @attr[:href].to_s
        @owner_class.module_eval %Q{
          def #{VIEW_PREFIX}_#{method}
            a({:href=>"#{href}"}.update(#{element_attr_str(method)}))
          end
        }
      end
=begin
      def define_view_method_body(clazz, method)
        clazz.module_eval %Q{
          def #{VIEW_PREFIX}_#{method}
            a({:href=>#{method}}.update(#{element_attr_str(method)}))
          end
        }
      end
=end
      register
    end

    class Textbox < Value; register; end
    class ActionTextbox < ActionValue; register; end

    class Checkbox < Base
      def do_inject_parameter(obj, attr, param)
        if obj.respond_to?("#{attr}=")
          if param[attr]
            obj.__send__("#{attr}=", true)
          else
            # ⓚpfalseɂ̂ŃoỎ\
            obj.__send__("#{attr}=", false)
          end
          puts "INJECTED #{attr} => id:#{obj.__id__}" if $DEBUG
        end
      end
      def define_view_method_body(clazz, method)
        clazz.module_eval %Q{
          def #{VIEW_PREFIX}_#{method}
            if #{method} == true
              a({:checked=>'checked'}.update(#{element_attr_str(method)}))
            else
              a({:checked=>nil}.update(#{element_attr_str(method)}))
            end
          end
        }
      end
      register
    end
    
    class ActionCheckbox < Checkbox
      include ActionReplace
      register
    end
    
    class Radio < Base
      def self.spec_option
        { :replace_attr => true,
          :radio_selection_element => [:checked, []] }
      end
      register
    end
    class ActionRadio < Radio
      include ActionReplace
      register
    end

    class AbstractChoiceDynamic < Base
      def choice_name
        'choice_name'
      end
      def define_view_method
        method = @attr[:name].to_s
        item_id = @attr[:item_id].to_s
        items_name = @attr[:items].to_s
        text_name = (@attr[:text] || :text).to_s
        value_name = (@attr[:value] || :value).to_s
        @owner_class.module_eval %Q{
          def #{VIEW_PREFIX}_#{method}(m)
            m.#{VIEW_PREFIX}_#{method}(:name=>"#{method}_\#{__id__}") do |mm|
              method2 = mm::Methods.first
              items = @#{items_name}
              items = [items] unless items.instance_of?(Array)
              items.each do |it|
                text = "#{text_name}".to_sym
                value = "#{value_name}".to_sym
                if it.instance_of?(Hash)
                  if #{method} == it[value]
                    mm.__send__(method2, it[text], {:value=>it[value], :'#{choice_name}'=>'#{choice_name}'}.update(#{element_attr_str(method)}))
                  else
                    mm.__send__(method2, it[text], {:value=>it[value], :'#{choice_name}'=>nil}.update(#{element_attr_str(method)}))
                  end
                else
                  if #{method} == it.__send__(value)
                    mm.__send__(method2, it.__send__(text), {:value=>it.__send__(value), :'#{choice_name}'=>'#{choice_name}'}.update(#{element_attr_str(method)}))
                  else
                    mm.__send__(method2, it.__send__(text), {:value=>it.__send__(value), :'#{choice_name}'=>nil}.update(#{element_attr_str(method)}))
                  end
                end
              end
            end
          end
        }
      end
    end

    class RadioDynamic < AbstractChoiceDynamic
      def self.spec_option
        { :replace_attr => true }
      end
      def choice_name
        'checked'
      end
      register
    end

    class ActionRadioDynamic < RadioDynamic
      include ActionReplace
      register
    end

    class Select < Base
      def self.spec_option
        { :replace_attr => true,
          :selection_element => [:selected, []] }
      end
      register
    end
    class ActionSelect < Select
      include ActionReplace
      register
    end

    class SelectDynamic < AbstractChoiceDynamic
      def self.spec_option
        { :replace_attr => true }
      end
      def choice_name
        'selected'
      end
      register
    end

    class ActionSelectDynamic < SelectDynamic
      include ActionReplace
      register
    end

    # ftHgr[\bho^
    Annotation::ScanHandler.register_before do |const|
      if const.instance_of?(Class)
        clazz = const
        read_attrs = clazz.instance_methods(true) - Object.instance_methods(true)
#        read_attrs.reject! {|m| m[-1, 1] == '='}
        read_attrs.reject! {|m| /\W/ =~ m}
        puts "Class #{clazz.name} readers: #{read_attrs.inspect}" if $DEBUG
        # ftHgr[\bh
        read_attrs.each do |method|
          clazz.module_eval %Q{
            def #{VIEW_PREFIX}_#{method}
              #{method}
            end
          }
        end
      end
    end

  end
end
